#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// Ranking ////////

void botInfo::init_rank(Player *p)
{
	Sint32 points = p->score.flagPoints + p->score.killPoints;

	for (int i = 0; i < NUM_RANKS; ++i)
	{
		if (points < setts.RankPoints[i])
		{
			--i;

			set_tag(p, 0, i);

			return;
		}
	}

	set_tag(p, 0, NUM_RANKS - 1);
}

void botInfo::check_rank(Player *p)
{
	int oldrank = get_tag(p, 0);

	Sint32 score = p->score.flagPoints + p->score.killPoints;

	for (int i = 0; i < NUM_RANKS; ++i)
	{
		if (score < setts.RankPoints[i])
			break;
	}

	int rank = i - 1;

	set_tag(p, 0, rank);

	if (rank == oldrank)
		return;

	if (rank > oldrank)
	{
		String s;
		s += "You have be promoted to ";
		s += setts.RankName[rank];

		if (rank == NUM_RANKS - 1)
		{
			s += ", the highest honor you may receive.";

			String m;
			m += "*arena A player has achieved the highest rank one may receive. Please greet ";
			m += setts.RankName[rank];
			m += " ";
			m += p->name;
			sendPublic(SND_Ooooo, m);
		}
		else
		{
			s += ". ";
			s += setts.RankPoints[rank + 1];
			s += " points for the next rank.";
		}

		sendPrivate(p, s);

		s = "*warn You may only use ships: ";

		for (int i = 0; i < NUM_SHIPS; ++i)
		{
			if ((rank + 1) >= setts.ShipLevels[i])
			{
				s += setts.ShipName[i];
				s += "#";
				s += i + 1;
				s += ", ";
			}
		}

		s += "until your next promotion.";

		sendPrivate(p, SND_Hallellula, s);
	}
	else if ((rank + 1) < setts.ShipLevels[p->ship])
	{
		for (int i = NUM_SHIPS - 1; i >= 0; --i)
		{
			if ((rank + 1) >= setts.ShipLevels[i])
			{
				String s;
				s += "*setship ";
				s += (i + 1);
				sendPrivate(p, s);
				break;
			}
		}
	}
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
			for (int i = 0; i < 4; ++i)
				--countdown[i];
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}

			_listnode <Player> *parse = playerlist->head;

			while (parse)
			{
				Player *p = parse->item;

				init_rank(p);

				parse = parse->next;
			}
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			killTags();
			reset_scores = false;
			energy_on = false;
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];

			_listnode <Player> *parse = playerlist->head;

			while (parse)
			{
				Player *p = parse->item;

				if (p->team == team)
					check_rank(p);

				parse = parse->next;
			}
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];

			_listnode <Player> *parse = playerlist->head;

			while (parse)
			{
				Player *p = parse->item;

				if (p->team == team)
					check_rank(p);

				parse = parse->next;
			}
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
			{
				p[i] = (Player*)event.p[i];

				if (p[i])
					check_rank(p[i]);
			}
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];

			_listnode <Player> *parse = playerlist->head;

			while (parse)
			{
				Player *p = parse->item;

				if (p->team == team)
				check_rank(p);

				parse = parse->next;
			}
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];

			init_rank(p);

			if (reset_scores)
			{
				sendPrivate(p, "*scorereset");
			}

			if (energy_on)
			{
				sendPrivate(p, "*energy");
			}

			int rank = get_tag(p, 0);

			String s;

			s += "Welcome back ";
			s += setts.RankName[rank];
			s += " ";
			s += p->name;
			s += ".";

			if (rank != NUM_RANKS - 1)
			{
				s += " ";
				s += setts.RankPoints[rank + 1];
				s += " points for the next rank.";
			}

			sendPrivate(p, SND_MessageAlarm, s);

			s = "*warn You may only use these ships: ";

			for (int i = 0; i < NUM_SHIPS; ++i)
			{
				if ((rank + 1) >= setts.ShipLevels[i])
				{
					s += setts.ShipName[i];
					s += "#";
					s += i + 1;
					s += ", ";
				}
			}

			s += "until your next promotion.";

			sendPrivate(p, SND_MessageAlarm, s);

			if (p->ship == SHIP_Spectator)
				break;

			if ((rank + 1) < setts.ShipLevels[p->ship])
			{
				for (int i = NUM_SHIPS - 1; i >= 0; --i)
				{
					if ((rank + 1) >= setts.ShipLevels[i])
					{
						String s;
						s += "*setship ";
						s += (i + 1);
						sendPrivate(p, s);
						break;
					}
				}
			}
			else
			{
				for (int i = 0; i < NUM_PRIZES; ++i)
				{
					if (setts.PrizeLevels[i] > 0)
					if ((rank + 1) >= setts.PrizeLevels[i])
					{
						String s;
						s += "*prize #";
						s += upgrade_prizes[i];
						sendPrivate(p, s);
					}
				}
			}
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];

			if (get_tag(p, 1) == 31337)
			for (int i = 0; i < NUM_PRIZES; ++i)
			{
				if (setts.PrizeLevels[i] > 0)
				if ((get_tag(p, 0) + 1) >= setts.PrizeLevels[i])
				{
					String s;
					s += "*prize #";
					s += upgrade_prizes[i];
					sendPrivate(p, s);
				}
			}

			set_tag(p, 1, 0);
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];

			check_rank(k);

			set_tag(p, 1, 31337);
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];

			check_rank(p);
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldship = (Uint16)event.p[1];

			if (p->ship == SHIP_Spectator)
				break;

			if ((get_tag(p, 0) + 1) < setts.ShipLevels[p->ship])
			{
				for (int i = NUM_SHIPS - 1; i >= 0; --i)
				{
					if ((get_tag(p, 0) + 1) >= setts.ShipLevels[i])
					{
						String s;
						s += "*setship ";
						s += (i + 1);
						sendPrivate(p, s);
						break;
					}
				}
			}
			else
			{
				for (int i = 0; i < NUM_PRIZES; ++i)
				{
					if (setts.PrizeLevels[i] > 0)
					if ((get_tag(p, 0) + 1) >= setts.PrizeLevels[i])
					{
						String s;
						s += "*prize #";
						s += upgrade_prizes[i];
						sendPrivate(p, s);
					}
				}
			}
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];

			if (p->ship == SHIP_Spectator)
				break;

			for (int i = 0; i < NUM_PRIZES; ++i)
			{
				if (setts.PrizeLevels[i] > 0)
				if ((get_tag(p, 0) + 1) >= setts.PrizeLevels[i])
				{
					String s;
					s += "*prize #";
					s += upgrade_prizes[i];
					sendPrivate(p, s);
				}
			}
		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];
			killTags(p);
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major = HIWORD(*(int*)&event.p[0]);
			int minor = LOWORD(*(int*)&event.p[0]);
			callback = (CALL_COMMAND)event.p[1];
			playerlist = (CALL_PLIST)event.p[2];
			flaglist = (CALL_FLIST)event.p[3];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
				tell(makeEcho("DLL plugin connected."));
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Tag emulation ////////

int botInfo::get_tag(Player *p, int index)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
			return tag->data;

		parse = parse->next;
	}

	return 0;
}

void botInfo::set_tag(Player *p, int index, int data)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
		{
			tag->data = data;
			return;
		}

		parse = parse->next;
	}

	tag = new PlayerTag;
	tag->p = p;
	tag->index = index;
	tag->data = data;
	taglist.append(tag);
}

void botInfo::killTags(Player *p)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;
		parse = parse->next;

		if (tag->p != p) continue;

		taglist.kill(tag);
	}
}

void botInfo::killTags()
{
	taglist.clear();
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
